//============================================================================
// Name        : led.c
// Author      : K. McIntire
// Version     :
// Copyright   : Red Lion Controls, 2022
//               Red Lion and the Red Lion logo are registered trademarks of Red Lion Controls, Inc.
//               All other company and product names are trademarks of their respective owners.
// Description : This file contains the support for controlling LEDs
//               from the Redboot bootloader
//
//============================================================================
#include <redboot.h>
#include <led.h>

#include <stdio.h>                                                   // sscanf needed

#define TRASH_LEN       50                                           // Used for sscanf trash detection
#define CFG_BUF_SIZE    100                                          // Size of Configuration buffer

RedBoot_cmd("led",
            "Control one or more LEDs",
            "[-y] <yellow LEDs> [-r] <red LEDs> [-g] <green LEDs> [-n] <off LEDs>",
            do_led_ctrl
    );

unsigned char ledState[MAX_LEDS];                                    // Tracks desired state of the LED

//----------------------------------------------------------------------------
// Name:    do_led_ctrl
//
// Desc:    Code that implements the Redboot "led" command
//
// Args:    int argc     - the number of command line arguments
//          char *argv[] - tokenized command line arguments
//
// Return:  void
//
//----------------------------------------------------------------------------
void do_led_ctrl(int argc, char *argv[])
{
    int             ndx = 1,                                        // argv index
                    errors = FALSE,                                 // TRUE when there is an unrecoverable command line problem
                    cygStatus;                                      // Driver Interface call status

    cyg_io_handle_t drvrHndl;                                       // Handle to the LEDs driver

    cyg_uint32      bufSize;                                        // Number of bytes in transfer buffer

    memset(ledState, 0, MAX_LEDS);                                  // Default to all LEDs turned off

    while ( !errors && (ndx < argc) )                               // Loop through command line args
    {
        if ( !strcmp(argv[ndx], LED_YELLOW_CMND_ARG) )              // Yellow LED argument
        {
            errors = !initializeLedState(argc,                      // Initialize one or more LEDs to yellow
                                        argv,
                                        ndx,
                                        LED_YELLOW);
        }
        else if ( !strcmp(argv[ndx], LED_GREEN_CMND_ARG) )          // Green LED argument
        {
            errors = !initializeLedState(argc,                      // Initialize one or more LEDs to green
                                        argv,
                                        ndx,
                                        LED_GREEN);
        }
        else if ( !strcmp(argv[ndx], LED_RED_CMND_ARG) )            // Red LED argument
        {
            errors = !initializeLedState(argc,                      // Initialize one or more LEDs to red
                                        argv,
                                        ndx,
                                        LED_RED);
        }
        else if ( !strcmp(argv[ndx], LED_OFF_CMND_ARG) )            // Off LED argument
        {
            errors = !initializeLedState(argc,                      // Initialize one or more LEDs to off
                                        argv,
                                        ndx,
                                        LED_OFF);
        }
        else                                                        // Unsupported command line argument
        {
            errors = TRUE;

            err_printf("Unrecognized 'led' command argument: [%s]\n", argv[ndx]);
        }

        if ( !errors ) ndx += 2;                                    // Go to next command line argument
    }

    if ( !errors )                                                  // LED state is OK
    {
        cygStatus = cyg_io_lookup("/dev/led0",                      // Open interface to LED driver
                                  &drvrHndl);

        if ( cygStatus != ENOERR )                                  // Problem communicating with driver
        {
            errors = TRUE;

            err_printf("Could not access /dev/led0 driver [%d]\n", cygStatus );
        }
    }

    if ( !errors )                                                  // Interface to Driver is OK
    {
        bufSize = MAX_LEDS;

        cygStatus = cyg_io_write(drvrHndl,                          // Set the LEDS
                                 ledState,
                                 &bufSize);

        if ( cygStatus != ENOERR )                                  // Problem Setting the LEDs
        {
            errors = TRUE;

            err_printf("Could not set the state of the LEDs [%d]\n", cygStatus );
        }
    }

    if ( !errors ) getDriverCfg(drvrHndl);                          // Output driver configuration
}

//----------------------------------------------------------------------------
// Name:    initializeLedState
//
// Desc:    Initializes one or more LEDs to the specified state.
//
// Args:    int argc               - the number of command line arguments
//          char *argv[]           - tokenized command line arguments
//          int ndx                - current command line argument
//          unsigned char ledState - State of the specified LED(s)
//
// Return:  Zero if there was an unrecoverable command line error.
//          Otherwise, non-zero.
//
//----------------------------------------------------------------------------
int initializeLedState(int argc,                                    // Total number of command line arguments
                       char *argv[],                                // command line arguments
                       int ndx,                                     // current command line argument
                       unsigned char newLedState)                   // LED state to set
{
    int status = TRUE,                                              // Default to successful status
        lowRange,                                                   // Low end of LED index range
        hiRange,                                                    // High end of LED index range
        ledNdx,                                                     // Current LED index
        cnt1 = 0,                                                   // sscanf1 count
        cnt2 = 0;                                                   // sscanf2 count

    char trash[TRASH_LEN];                                          // sscanf trash detection

    if ( (ndx + 1) >= argc )                                        // Dangling command line argument
    {
        status = FALSE;

        err_printf("Expected command line argument after [%s] is missing\n", argv[ndx]);
    }
    else if ( newLedState > LED_YELLOW )                            // Unsupported LED state
    {
        status = FALSE;

        err_printf("Unsupported LED state [%d] specified. Programming error\n", ledState);
    }
    else if ( (cnt1 = sscanf(argv[ndx + 1],                         // LED index range check
                             "%d-%d%s",
                             &lowRange,
                             &hiRange,
                             trash)) >= 2 )
    {
        if ( hiRange < lowRange )                                   // Flip the range limits
        {
            ledNdx = lowRange;
            lowRange = hiRange;
            hiRange = ledNdx;
        }

        if ( lowRange < 0 ) lowRange = 0;                           // Reset low end of range

        if ( hiRange >= MAX_LEDS ) hiRange = MAX_LEDS - 1;          // Reset high end of range

        for ( ledNdx = lowRange;                                    // Loop through specified LED range
              ledNdx <= hiRange;
              ledNdx++ )
        {
            ledState[ledNdx] = newLedState;
        }
    }
    else if ( (cnt2 = sscanf(argv[ndx + 1],                         // Single LED index check
                             "%d%s",
                             &ledNdx,
                             trash)) >= 1 )
    {
        if ( (ledNdx >= 0) && (ledNdx < MAX_LEDS) )                 // Good LED index was specified
        {
            ledState[ledNdx] = newLedState;
        }
    }
    else                                                            // Unrecognizable LED index
    {
        status = FALSE;

        err_printf("Unrecognizable LED index specified [%s] after %s argument\n",
                   argv[ndx + 1],
                   argv[ndx]);
    }

    return status;
}

//----------------------------------------------------------------------------
// Name:    getDriverCfg
//
// Desc:    Gets and prints the LEDs driver configuration.
//
// Args:    hndl  - opaque handle to the driver
//
// Return:  void
//
//----------------------------------------------------------------------------
void getDriverCfg(cyg_io_handle_t hndl)
{
    int             cygStatus;                                      // Driver call status

    cyg_uint32      value,                                          // Configuration value
                    valueLen = sizeof(value);                       // Bytes in Configuration value

    if ( (cygStatus = cyg_io_get_config(hndl,                       // Get sgpio clock pin
                                        LED_KEY_CLOCK_PIN,
                                        &value,
                                        &valueLen)) == ENOERR )
    {
        err_printf("DRIVER CFG: Bitmask for gpio pin for sgpio clock [0x%08lx]\n", value);
    }
    else
    {
        err_printf("DRIVER CFG: Could not get gpio pin for sgpio clock [%d]\n", cygStatus);
    }

    if ( (cygStatus = cyg_io_get_config(hndl,                       // Get sgpio Data Out pin
                                        LED_KEY_DO_PIN,
                                        &value,
                                        &valueLen)) == ENOERR )
    {
        err_printf("DRIVER CFG: Bitmask for gpio pin for sgpio data out [0x%08lx]\n", value);
    }
    else
    {
        err_printf("DRIVER CFG: Could not get gpio pin for sgpio data out [%d]\n", cygStatus);
    }

    if ( (cygStatus = cyg_io_get_config(hndl,                       // Get sgpio Load Data pin
                                        LED_KEY_LD_PIN,
                                        &value,
                                        &valueLen)) == ENOERR )
    {
        err_printf("DRIVER CFG: Bitmask for gpio pin for sgpio load data [0x%08lx]\n", value);
    }
    else
    {
        err_printf("DRIVER CFG: Could not get gpio pin for sgpio load data [%d]\n", cygStatus);
    }

    if ( (cygStatus = cyg_io_get_config(hndl,                       // Get Enable pin
                                        LED_KEY_EN_PIN,
                                        &value,
                                        &valueLen)) == ENOERR )
    {
        err_printf("DRIVER CFG: Bitmask for gpio pin Enable [0x%08lx]\n", value);
    }
    else
    {
        err_printf("DRIVER CFG: Could not get gpio pin Enable [%d]\n", cygStatus);
    }

    if ( (cygStatus = cyg_io_get_config(hndl,                       // Output Enable Register
                                        LED_KEY_OE_RGSTR,
                                        &value,
                                        &valueLen)) == ENOERR )
    {
        err_printf("DRIVER CFG: Output Enable Register [0x%08lx]\n", value);
    }
    else
    {
        err_printf("DRIVER CFG: Could not get Output Enable Register [%d]\n", cygStatus);
    }

    if ( (cygStatus = cyg_io_get_config(hndl,                       // Output Enable Register Value
                                        LED_KEY_OE_RGSTR_VAL,
                                        &value,
                                        &valueLen)) == ENOERR )
    {
        err_printf("DRIVER CFG: Output Enable Register Value [0x%08lx]\n", value);
    }
    else
    {
        err_printf("DRIVER CFG: Could not get Output Enable Register Value [%d]\n", cygStatus);
    }

    if ( (cygStatus = cyg_io_get_config(hndl,                       // Output Set Register
                                        LED_KEY_SET_RGSTR,
                                        &value,
                                        &valueLen)) == ENOERR )
    {
        err_printf("DRIVER CFG: Output Set Register [0x%08lx]\n", value);
    }
    else
    {
        err_printf("DRIVER CFG: Could not get Output Set Register [%d]\n", cygStatus);
    }

    if ( (cygStatus = cyg_io_get_config(hndl,                       // Output Clear Register
                                        LED_KEY_CLR_RGSTR,
                                        &value,
                                        &valueLen)) == ENOERR )
    {
        err_printf("DRIVER CFG: Output Clear Register [0x%08lx]\n", value);
    }
    else
    {
        err_printf("DRIVER CFG: Could not get Output Clear Register [%d]\n", cygStatus);
    }

    if ( (cygStatus = cyg_io_get_config(hndl,                       // Output Register
                                        LED_KEY_OUT_RGSTR,
                                        &value,
                                        &valueLen)) == ENOERR )
    {
        err_printf("DRIVER CFG: Output Register [0x%08lx]\n", value);
    }
    else
    {
        err_printf("DRIVER CFG: Could not get Output Register [%d]\n", cygStatus);
    }

    if ( (cygStatus = cyg_io_get_config(hndl,                       // Output Register Value
                                        LED_KEY_OUT_RGSTR_VAL,
                                        &value,
                                        &valueLen)) == ENOERR )
    {
        err_printf("DRIVER CFG: Output Register Value [0x%08lx]\n", value);
    }
    else
    {
        err_printf("DRIVER CFG: Could not get Output Register Value [%d]\n", cygStatus);
    }

    if ( (cygStatus = cyg_io_get_config(hndl,                       // LED Count
                                        LED_KEY_LED_CNT,
                                        &value,
                                        &valueLen)) == ENOERR )
    {
        err_printf("DRIVER CFG: LED Count [%d]\n", value);
    }
    else
    {
        err_printf("DRIVER CFG: Could not get LED Count [%d]\n", cygStatus);
    }

    if ( (cygStatus = cyg_io_get_config(hndl,                       // Least Significant Shifts
                                        LED_KEY_SHIFTS_0,
                                        &value,
                                        &valueLen)) == ENOERR )
    {
        err_printf("DRIVER CFG: Least Significant Bit Shifts [0x%08lx]\n", value);
    }
    else
    {
        err_printf("DRIVER CFG: Could not get Least Significant Bit Shifts [%d]\n", cygStatus);
    }

    if ( (cygStatus = cyg_io_get_config(hndl,                       // Middlemost Significant Shifts
                                        LED_KEY_SHIFTS_1,
                                        &value,
                                        &valueLen)) == ENOERR )
    {
        err_printf("DRIVER CFG: Middlemost Significant Bit Shifts [0x%08lx]\n", value);
    }
    else
    {
        err_printf("DRIVER CFG: Could not get Middlemost Significant Bit Shifts [%d]\n", cygStatus);
    }

    if ( (cygStatus = cyg_io_get_config(hndl,                       // Most Significant Shifts
                                        LED_KEY_SHIFTS_2,
                                        &value,
                                        &valueLen)) == ENOERR )
    {
        err_printf("DRIVER CFG: Most Significant Bit Shifts [0x%08lx]\n", value);
    }
    else
    {
        err_printf("DRIVER CFG: Could not get Most Significant Bit Shifts [%d]\n", cygStatus);
    }

    return;
}
// EOF led.c
