//==========================================================================
//
//      mscc_simage.c
//
//      Support code/data for MSCC signed image
//
//==========================================================================
// ####ECOSGPLCOPYRIGHTBEGIN####                                            
// -------------------------------------------                              
// This file is part of eCos, the Embedded Configurable Operating System.   
// Copyright (C) 2006 Free Software Foundation, Inc.                        
//
// eCos is free software; you can redistribute it and/or modify it under    
// the terms of the GNU General Public License as published by the Free     
// Software Foundation; either version 2 or (at your option) any later      
// version.                                                                 
//
// eCos is distributed in the hope that it will be useful, but WITHOUT      
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or    
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License    
// for more details.                                                        
//
// You should have received a copy of the GNU General Public License        
// along with eCos; if not, write to the Free Software Foundation, Inc.,    
// 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.            
//
// As a special exception, if other files instantiate templates or use      
// macros or inline functions from this file, or you compile this file      
// and link it with other works to produce a work based on this file,       
// this file does not by itself cause the resulting work to be covered by   
// the GNU General Public License. However the source code for this file    
// must still be made available in accordance with section (3) of the GNU   
// General Public License v2.                                               
//
// This exception does not invalidate any other reasons why a work based    
// on this file might be covered by the GNU General Public License.         
// -------------------------------------------                              
// ####ECOSGPLCOPYRIGHTEND####                                              
//==========================================================================
//#####DESCRIPTIONBEGIN####
//
// Author(s):    Lars Povlsen
// Contributors: 
// Date:         2015-12-08
// Purpose:      
// Description:  
//              
// This code is part of RedBoot (tm).
//
//####DESCRIPTIONEND####
//
//==========================================================================

#include <redboot.h>
#include <cyg/infra/cyg_ass.h>         // assertion macros

#include "mscc_simage.h"

#if defined(CYGPKG_CRYPT_LIBTOMCRYPT)
#include <cyg/libtomcrypt/tomcrypt.h>
#endif  /* CYGPKG_CRYPT_LIBTOMCRYPT */

#define SHA1_HASH_SIZE 20

#define T_N(fmt, ...) // diag_printf("noise: "   fmt "\n", ## __VA_ARGS__)
#define T_D(fmt, ...) // diag_printf("debug: "   fmt "\n", ## __VA_ARGS__)
#define T_I(fmt, ...) // diag_printf("info: "    fmt "\n", ## __VA_ARGS__)
#define T_W(fmt, ...) diag_printf("warning: " fmt "\n", ## __VA_ARGS__)
#define T_E(fmt, ...) diag_printf("error: "   fmt "\n", ## __VA_ARGS__)

typedef struct {
    const cyg_uint8 *buf_ptr;
    size_t    buf_len;
} bufptr_t;

#if defined(CYGPKG_CRYPT_LIBTOMCRYPT)
static bool mac_check(int algo, const cyg_uint8 *buf, size_t len, const cyg_uint8 *digest)
{
    cyg_uint8 calc_hash[SHA1_HASH_SIZE];
    unsigned long digest_length = SHA1_HASH_SIZE;
    if(hmac_memory(algo, (const cyg_uint8*) SIMAGE_KEY, strlen(SIMAGE_KEY), 
                   buf, len, calc_hash, &digest_length) != CRYPT_OK ||
       digest_length != SHA1_HASH_SIZE ||
       digest_length > SIMAGE_DIGLEN) {
        T_E("MAC calculation error");
        return false;
    }
    return (memcmp(digest, calc_hash, digest_length) == 0);
}
#endif

static const cyg_uint8 *
bufsearch(const cyg_uint8 *haystack, size_t length, const cyg_uint8 *needle, size_t nlength)
{
    const cyg_uint8 *end = haystack+length;
    while(haystack < end &&
          (haystack = (const cyg_uint8*)memchr(haystack, needle[0], end - haystack))) {
        if(nlength <= (size_t)(end - haystack) &&
           memcmp(haystack, needle, nlength) == 0)
            return haystack;
        haystack++;
    }
    return NULL;
}

static cyg_uint32 get_dword(const cyg_uint8 *ptr)
{
    return (cyg_uint32) ((ptr[0]) + 
                  (ptr[1] << 8) +
                  (ptr[2] << 16) +
                  (ptr[3] << 24));
}

static inline void buf_init(bufptr_t *b, const cyg_uint8 *ptr, size_t len)
{
    b->buf_ptr = ptr;
    b->buf_len = len;
}

static inline void buf_skip(bufptr_t *b, size_t n)
{
    if(n <= b->buf_len) {
        b->buf_ptr += n;
        b->buf_len -= n;
    } else {
        b->buf_ptr += b->buf_len;
        b->buf_len = 0;
    }
}

static cyg_uint32 get_dword_advance(bufptr_t *b)
{
    cyg_uint32 val = get_dword(b->buf_ptr);
    buf_skip(b, sizeof(cyg_uint32));
    return val;
}

static void simage_init(simage_t *simg, const cyg_uint8 *image, size_t img_len)
{
    memset(simg, 0, sizeof(*simg));
    simg->img_ptr = image;
    simg->img_len = img_len;
}

static bool simage_validate_hash(simage_t *simg, const cyg_uint8 *p)
{
    const cyg_uint8 *ptrs = p;
    ptrs += SIMAGE_SIGLEN;      /* Skip Signature */
    simg->file_len = get_dword(ptrs); ptrs += 4;
    memcpy(simg->file_digest, ptrs, SIMAGE_DIGLEN); ptrs += SIMAGE_DIGLEN;
    simg->trailer_len = get_dword(ptrs); ptrs += 4;
    memcpy(simg->trailer_digest, ptrs, SIMAGE_DIGLEN); ptrs += SIMAGE_DIGLEN;
    simg->trailer_ptr = ptrs;
    if(simg->file_len < simg->img_len &&
       (simg->trailer_ptr + simg->trailer_len) < (simg->img_ptr + simg->img_len)) {
#if defined(CYGPKG_CRYPT_LIBTOMCRYPT)
        int algo = register_hash (&sha1_desc);
        if(algo < 0) {
            T_E("Unable to register SHA1 hash function, configuration error");
            return false;
        }
        if(!mac_check(algo, simg->img_ptr, simg->file_len, simg->file_digest)) {
            T_W("File MAC error");
            return false;
        }
        if(!mac_check(algo, simg->trailer_ptr, simg->trailer_len, simg->trailer_digest)) {
            T_W("Trailer MAC error");
            return false;
        }
#endif  /* CYGPKG_CRYPT_LIBTOMCRYPT */
        T_D("Image validated, now parsing");
        return true;
    }
    T_D("Ptr consistency error: lengths (file %zu, img %zu), img %p, trailer(%p , %zu)",
        simg->file_len, simg->img_len, simg->img_ptr, simg->trailer_ptr, simg->trailer_len);
    return false;
}

static const cyg_uint8 *simage_search_trailer(simage_t *simg, bufptr_t *b)
{
    const cyg_uint8 *hit;
    if(b->buf_len &&
       (hit = bufsearch(b->buf_ptr, b->buf_len, (const cyg_uint8*)SIMAGE_SIG, SIMAGE_SIGLEN))) {
        buf_skip(b, hit - b->buf_ptr); /* Skip down */
        T_D("Found signature at %p, buffer left %zu", hit, b->buf_len);
        return b->buf_ptr;
    }
    T_D("No signature match");
    return NULL;
}

static int simage_get_trailer(simage_t *simg, bufptr_t *b)
{
    cyg_uint32 backp;
    const cyg_uint8 *p;
    /* Try to use the backpointer, if found */
    p = (b->buf_ptr + b->buf_len); /* End of buffer */
    if(get_dword(p - 2*sizeof(cyg_uint32)) == SIMAGE_TCOOKIE &&
       (backp = get_dword(p - 1*sizeof(cyg_uint32))) < b->buf_len &&
       memcmp(p - backp, SIMAGE_SIG, SIMAGE_SIGLEN) == 0) {
        T_D("Signature found with backpointer: %d", backp);
        p -= backp;
        return simage_validate_hash(simg, p) ? 0 : -1;
    } else {
        /* Must search from start */
        while((p = simage_search_trailer(simg, b))) {
            if(simage_validate_hash(simg, p))
                return 0;
            else 
                /* Try for other matches */
                buf_skip(b, SIMAGE_SIGLEN);
        }
    }

    /* All failed */
    return -1;
}

int simage_parse_tlvs(simage_t *simg)
{
    bufptr_t buf;
    buf_init(&buf, simg->trailer_ptr, simg->trailer_len);
    while(buf.buf_len > 3*sizeof(cyg_uint32)) {
        cyg_uint32 type, id, dlen;
        type = get_dword_advance(&buf);
        id = get_dword_advance(&buf);
        dlen = get_dword_advance(&buf);
        if(id > 0 && id < SIMAGE_TLV_COUNT) {
            simage_tlv_t *tlv = &simg->tlv[id];
            T_N("type %d, id %d dlen %d", type, id, dlen);
            switch(type) {
            case SIMAGE_TLV_TYPE_DWORD:
                tlv->valid = true;
                tlv->type = type;
                tlv->dw_value = get_dword(buf.buf_ptr);
                T_D("Dword id %d value %d", id, tlv->dw_value);
                break;
            case SIMAGE_TLV_TYPE_STRING:
                tlv->valid = true;
                tlv->type = type;
                tlv->str_len = dlen;
                tlv->str_ptr = buf.buf_ptr;
                T_D("Str id %d value %s", id, buf.buf_ptr);
                break;
            case SIMAGE_TLV_TYPE_BINARY:
                tlv->valid = true;
                tlv->type = type;
                tlv->str_len = dlen;
                tlv->str_ptr = buf.buf_ptr;
                T_D("Bin id %d len %u", id, dlen);
                break;
            default:
                T_I("Skip invalid type: type %d, id %d dlen %d", type, id, dlen);
            }
        } else {
            T_I("Skip invalid id: %d, id %d dlen %d", type, id, dlen);
        }
        buf_skip(&buf, dlen);
    }
    /* Always succeeds, even if skipped TLV's */
    return 0;
}

int simage_parse(simage_t *simg, const cyg_uint8 *image, size_t img_len)
{
    bufptr_t buf;
    int rc;
    simage_init(simg, image, img_len);
    buf_init(&buf, image, img_len);
    if((rc = simage_get_trailer(simg, &buf)) != 0)
        return rc; 
    return simage_parse_tlvs(simg);
}

bool simage_get_tlv_dword(const simage_t *simg, int id, cyg_uint32 *val)
{
    if(id > 0 && id < SIMAGE_TLV_COUNT) {
        const simage_tlv_t *tlv = &simg->tlv[id];
        if(tlv->valid && tlv->type == SIMAGE_TLV_TYPE_DWORD) {
            *val = tlv->dw_value;
            return true;
        }
    } 
    return false;
}

/****************************************************************************/
/*                                                                          */
/*  End of file.                                                            */
/*                                                                          */
/****************************************************************************/
