//==========================================================================
//
//      ramload.c
//
//      RedBoot memory copy (ramload) routine
//
//==========================================================================
// ####ECOSGPLCOPYRIGHTBEGIN####                                            
// -------------------------------------------                              
// This file is part of eCos, the Embedded Configurable Operating System.   
// Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 Free Software Foundation, Inc.
//
// eCos is free software; you can redistribute it and/or modify it under    
// the terms of the GNU General Public License as published by the Free     
// Software Foundation; either version 2 or (at your option) any later      
// version.                                                                 
//
// eCos is distributed in the hope that it will be useful, but WITHOUT      
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or    
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License    
// for more details.                                                        
//
// You should have received a copy of the GNU General Public License        
// along with eCos; if not, write to the Free Software Foundation, Inc.,    
// 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.            
//
// As a special exception, if other files instantiate templates or use      
// macros or inline functions from this file, or you compile this file      
// and link it with other works to produce a work based on this file,       
// this file does not by itself cause the resulting work to be covered by   
// the GNU General Public License. However the source code for this file    
// must still be made available in accordance with section (3) of the GNU   
// General Public License v2.                                               
//
// This exception does not invalidate any other reasons why a work based    
// on this file might be covered by the GNU General Public License.         
// -------------------------------------------                              
// ####ECOSGPLCOPYRIGHTEND####                                              
//==========================================================================
//#####DESCRIPTIONBEGIN####
//
// Author(s):    lpovlsen
// Contributors: 
// Date:         2013-10-22
// Purpose:      
// Description:  
//              
// This code is part of RedBoot (tm).
//
//####DESCRIPTIONEND####
//
//==========================================================================

#include <redboot.h>
#include <fis.h>
#include <flash_config.h>
#ifdef CYGPKG_COMPRESS_XZLIB
#include <cyg/crc/crc.h>
#include <vimage.h>
#endif

struct dec_out {
    CYG_ADDRESS dst_address;
    CYG_ADDRESS start_address;
    size_t decoded_length;
    const char *err_msg;
};

static void ramload_do_cmd(char *command, int no_exec)
{
    struct cmd *cmd;

    if (no_exec) {
         diag_printf("Execute the image with the following command:\n%s\n", command);
    } else {
        diag_printf("%s\n", command);
        if ((cmd = parse(&command, &argc, &argv[0])) != (struct cmd *)0) {
            (cmd->fun)(argc, argv);
        } else {
            diag_printf("Invalid command: %s\n", command);
        }
    }
}

static int ramload_setup(char *name, CYG_ADDRESS defaddr, struct dec_out *decode, unsigned long dst_address)
{
    struct fis_image_desc *img;

    memset(decode, 0, sizeof(*decode));
    if (dst_address) {
        decode->start_address = decode->dst_address = (CYG_ADDRESS)dst_address;
        diag_printf("Using explicit load address 0x%08x\n", decode->start_address);
    } else {
        if ((img = fis_lookup(name, NULL)) == (struct fis_image_desc *)0) {
            decode->start_address = decode->dst_address = defaddr;
            diag_printf("No image '%s' found, using default load address 0x%08x\n", name, decode->start_address);
        } else {
            decode->dst_address = img->mem_base;
            decode->start_address = img->entry_point;
        }

        if ((decode->dst_address) < ((CYG_ADDRESS)ram_start)) {
            diag_printf("Invalid destination address: %08x\n", decode->dst_address);
            decode->err_msg = "Invalid destination address";
            return -1;
        }
    }

    return 0;
}

static int ramload_gzip(unsigned char *src_address, struct dec_out *decode, int no_exec, unsigned long dst_address)
{
    char go_command[64];
    _pipe_t pipe, *p = &pipe;
    int err;

    if((err = ramload_setup("managed", (CYG_ADDRESS) 0x80040000, decode, dst_address)) != 0) {
        return err;
    }

    p->out_buf = (unsigned char*) decode->dst_address;
    p->out_max = p->out_size = -1;
    p->in_buf = src_address;
    p->in_avail = -1;

    err = (*_dc_init)(p);

    if (0 == err)
        err = (*_dc_inflate)(p);

    // Free used resources, do final translation of error value.
    err = (*_dc_close)(p, err);

    // Save error message
    decode->err_msg = p->msg;

    // Calculate decoded output length
    decode->decoded_length = ((CYG_ADDRESS)p->out_buf) - decode->dst_address;

    if(err == 0) {
        load_address     = decode->dst_address;
        load_address_end = decode->dst_address + decode->decoded_length;
        entry_address    = decode->start_address;
        diag_printf("Image loaded from 0x%08lx-0x%08lx, length %zu\n", 
                    load_address, load_address_end, decode->decoded_length);
        diag_sprintf(go_command, "go 0x%08x", decode->start_address);
        ramload_do_cmd(go_command, no_exec);
    }

    return err;
}

#ifdef CYGPKG_COMPRESS_XZLIB
static int decompress_xz(unsigned char *src_address, cyg_uint32 src_len, struct dec_out *decode)
{
    int err = xz_decompress_simple(src_address, src_len,
                                   (cyg_uint8*) decode->dst_address, (size_t) (src_address - decode->dst_address),
                                   &decode->decoded_length);
    if (!err) {
        diag_printf("XZ decompress to 0x%08x-0x%08x, length %zu\n", decode->dst_address, 
                    decode->dst_address + decode->decoded_length, decode->decoded_length);
    } else {
        decode->err_msg = xz_errmsg(err);
    }
    return err;
}

static int ramload_xz(unsigned char *src_address, struct dec_out *decode, int no_exec, unsigned long dst_address)
{
    char go_command[64];
    int err;

    if((err = ramload_setup("linux", (CYG_ADDRESS) 0x80100000, decode, dst_address)) != 0) {
        return err;
    }

    err = decompress_xz(src_address, workspace_end - src_address, decode);
    if (!err) {
        entry_address = decode->dst_address;
        diag_sprintf(go_command, "exec -w 2 0x%08x", decode->start_address);
        ramload_do_cmd(go_command, no_exec);
    } else {
        entry_address = (unsigned long)NO_MEMORY;
    }
    
    return err;
}

static int ramload_vimage(unsigned char *src_address, struct dec_out *decode, int no_exec, unsigned long dst_address)
{
    char go_command[64];
    struct firmware_vimage *fw = (struct firmware_vimage*) src_address;
    int err;

    if((err = ramload_setup("linux", (CYG_ADDRESS) 0x80100000, decode, dst_address)) != 0) {
        return err;
    }

    // Check FW header
    if ((err = vimage_hdrcheck(fw, &decode->err_msg))) {
        diag_printf("Invalid image: %s\n", decode->err_msg);
        return err;
    }

    diag_printf("Have vimage for %s/%s arch %d, length %u\n", 
                fw->machine, fw->soc_name, fw->soc_no, fw->imglen);

    err = decompress_xz(src_address + fw->kern_off, fw->kern_len, decode);
    if (!err) {
        entry_address = decode->dst_address;
        if (fw->kern_cmd_len) {
            diag_printf("Kernel command line: %s\n", fw->kern_cmd);
            linux_command_line = fw->kern_cmd;
        } else {
            linux_command_line = NULL;
        }
        // Prepare for possible root file system after kernel
        if (fw->root_len && fw->root_off) {
            struct squashfs_super_block *sb = (struct squashfs_super_block*) (src_address + fw->root_off);
            if (sb->s_magic == SQUASHFS_MAGIC &&
                sb->bytes_used_lo >= 0 && sb->bytes_used_hi == 0) {
                linux_ramdisk_start = (const cyg_uint8*) decode->dst_address + page_align(decode->decoded_length, 1024*1024);
                linux_ramdisk_size  = fw->root_len;
                memcpy((void*)linux_ramdisk_start, (void *)sb, linux_ramdisk_size);
                diag_printf("Moved RAM root file system from %p to %p, size %zd\n",
                            sb, linux_ramdisk_start, linux_ramdisk_size);
            } else {
                decode->err_msg = "Invalid rootfs data, must be squashfs";
                return -1;
            }
        }
        diag_sprintf(go_command, "exec -w 2 0x%08x", decode->start_address);
        ramload_do_cmd(go_command, no_exec);
    } else {
        entry_address = (unsigned long)NO_MEMORY;
    }
    
    return err;
}

static bool ramload_vimage_unified(mscc_firmware_vimage_t *fw, struct dec_out *decode, int no_exec, unsigned long dst_address)
{
    mscc_firmware_vimage_tlv_t tlv;
    u8 *tlvdata;
    char go_command[64];

    if(ramload_setup("linux", (CYG_ADDRESS) 0x80100000, decode, dst_address) != 0) {
        return false;
    }

    diag_printf("Have image for %s/%s arch %d, length %u\n", 
                fw->machine, fw->soc_name, fw->soc_no, fw->imglen);

    if ((tlvdata = mscc_vimage_find_tlv(fw, &tlv, MSCC_STAGE1_TLV_SIGNATURE))) {
        if (!mscc_vimage_validate(fw, &tlv, tlvdata)) {
            return false;
        }
    } else {
        diag_printf("Invalid image: Signature missing\n");
        return false;
    }

    if ((tlvdata = mscc_vimage_find_tlv(fw, &tlv, MSCC_STAGE1_TLV_METADATA))) {
        diag_write_string("Image metadata:\n");
        diag_write((char *)tlvdata, tlv.data_len);
        diag_write_string("\n");
    }

    if ((tlvdata = mscc_vimage_find_tlv(fw, &tlv, MSCC_STAGE1_TLV_KERNEL))) { // Kernel/stage1
        int err;
        diag_printf("Decompressing %d bytes...", tlv.data_len);
        err = xz_decompress_simple(tlvdata, tlv.data_len,
                                   (cyg_uint8*) decode->dst_address, (size_t) (tlvdata - decode->dst_address),
                                   &decode->decoded_length);
        diag_printf("done\n");
        if (!err) {
            entry_address = decode->start_address;
            diag_printf("Stage1: %p, length %d bytes\n", (unsigned char *)decode->dst_address, decode->decoded_length);
            // Update load address for optional initrd/rootfs
            decode->dst_address += page_align(decode->decoded_length, 1024*1024);
        } else {
            diag_printf("Decompression error: %s\n", xz_errmsg(err));
            return false;
        }
    } else {
        diag_printf("Invalid image: Kernel data missing\n");
        return false;
    }

    if ((tlvdata = mscc_vimage_find_tlv(fw, &tlv, MSCC_STAGE1_TLV_INITRD))) {
        struct squashfs_super_block *sb = (struct squashfs_super_block*) tlvdata;
        if (sb->s_magic == SQUASHFS_MAGIC &&
            sb->bytes_used_lo >= 0 && sb->bytes_used_hi == 0) {
            linux_ramdisk_start = (const cyg_uint8*) decode->dst_address;
            linux_ramdisk_size  = tlv.data_len;
            memmove((void*)decode->dst_address, tlvdata, tlv.data_len);
            diag_printf("Initrd: %p, length %d bytes\n", (void*)decode->dst_address, tlv.data_len);
        } else {
            diag_printf("Rootfs fails integrity tests, ignored\n");
        }
    }

    if ((tlvdata = mscc_vimage_find_tlv(fw, &tlv, MSCC_STAGE1_TLV_KERNEL_CMD))) {
        linux_command_line = (char *)tlvdata;
        diag_printf("Kernel command line: %s\n", linux_command_line);
    } else {
        linux_command_line = NULL;
    }

    // And start it off (perhaps)
    if (linux_command_line || linux_ramdisk_start) {
        diag_sprintf(go_command, "exec -w 2 0x%08x", decode->start_address);
    } else {
        diag_sprintf(go_command, "go 0x%08x", decode->start_address);
    }
    ramload_do_cmd(go_command, no_exec);

    return true;
}

#endif

static void do_ramload(int argc, char **argv)
{
    unsigned char *iptr;
    unsigned long src_address, dst_address;
    int err;
    struct dec_out decode;
    int no_exec = 0;
#ifdef CYGPKG_COMPRESS_XZLIB
    const char *errmsg;
#endif

    argc--, *argv++;    // Skip argv[0] == command
    while (argc && argv[0]) {
        if (argv[0][0] != '-') {
            break;
        }
        if (strcmp(argv[0], "--no-exec") == 0) {
            no_exec = 1;
        } else {
            diag_printf("Illegal option: %s\n", argv[0]);
      usage:
            diag_printf("Usage: ramload [--no-exec] [<src-address> [<dst-address>]]\n");
            diag_printf(" --no-exec:   Don't do exec/go after load\n");
            diag_printf(" src-address: Address of source data\n");
            diag_printf(" dst-address: Optional load (and entry) address\n");
            return;
        }
        argc--, *argv++;    // Next
    }

    if (argc > 2) {
        goto usage;
    }

    if (argc >= 1) {
        if (!parse_num(argv[0], &src_address, 0, 0)) {
            diag_printf("** Error: Invalid address: %s\n", argv[0]);
            goto usage;
        }
    } else {
        int memmap = 0;
        flash_get_config("linux_memmap", &memmap, CONFIG_INT);
        if (memmap == 0) {
            diag_printf("** Error: You must configure 'linux_memmap' with fconfig to omit source address\n");
            goto usage;
        }
        src_address = (CYG_ADDRESS) (ram_end - (memmap * (1024 * 1024)));
    }

    if (((CYG_ADDRESS)src_address) < ((CYG_ADDRESS)ram_start)) {
        diag_printf("** Error: Invalid source address: %08lx\n", src_address);
        goto usage;
    }

    if (argc == 2) {
        if (!parse_num(argv[1], &dst_address, 0, 0)) {
            diag_printf("** Error: Invalid address: %s\n", argv[1]);
            goto usage;
        }
        if (((CYG_ADDRESS)dst_address) < ((CYG_ADDRESS)ram_start)) {
            diag_printf("** Error: Invalid destination address: %08lx\n", dst_address);
            goto usage;
        }
    } else {
        dst_address = 0;
    }

    iptr = (unsigned char*) src_address;

    /* Check for gzip magic */
    if (iptr[0] == 0x1f && iptr[1] == 0x8b) {
        diag_printf("Decompressing from buffer start: %p\n", iptr);
        err = ramload_gzip(iptr, &decode, no_exec, dst_address);
    } else 
#ifdef CYGPKG_COMPRESS_XZLIB
    if (memcmp(iptr, xz_magic, sizeof(xz_magic)) == 0) {
        err = ramload_xz(iptr, &decode, no_exec, dst_address);
    } else if (vimage_is_legacy((cyg_uint32 *)iptr)) {
        err = ramload_vimage(iptr, &decode, no_exec, dst_address);
    } else if (mscc_vimage_hdrcheck((mscc_firmware_vimage_t *)iptr, &errmsg) == 0) {
        err = (ramload_vimage_unified((mscc_firmware_vimage_t *)iptr, &decode, no_exec, dst_address) == false);
    } else
#endif
    {
        /* Assume this is VTSS old-style image */
        iptr += 8;  /* CRC, len prefix */
        diag_printf("Decompressing from adjusted image start: %p\n", iptr);
        err = ramload_gzip(iptr, &decode, no_exec, dst_address);
    }

    if (0 != err && decode.err_msg) {
        entry_address = (CYG_ADDRESS) NO_MEMORY;
        diag_printf("Decompression error: %s\n", decode.err_msg);
    }
}

RedBoot_cmd("ramload", 
            "Relocate RAM image and boot it",
            "<ram-address> [--no-exec]",
            do_ramload
    );

#ifdef CYGPKG_COMPRESS_XZLIB

void
do_xz(int argc, char *argv[])
{
    struct option_info opts[2];
    unsigned long src, dst;
    bool src_set, dst_set;
    struct dec_out decode;

    init_opts(&opts[0], 's', true, OPTION_ARG_TYPE_NUM, 
              (void *)&src, (bool *)&src_set, "source address");
    init_opts(&opts[1], 'd', true, OPTION_ARG_TYPE_NUM, 
              (void *)&dst, (bool *)&dst_set, "destination address");
    if (!scan_opts(argc, argv, 1, opts, 2, 0, 0, "")) {
        return;
    }

    // Must have src and dst
    if (!src_set || !dst_set) {
        // try to use load_address for src 
        if (dst_set 
            && load_address >= (CYG_ADDRESS)ram_start 
            && load_address < load_address_end) {
            src = load_address;
            diag_printf("Decompressing from %p to %p\n",
			(void*)src, (void*)dst);
        } else {
            diag_printf("usage: xz -s <addr> -d <addr>\n");
            return;
        }
    }

    memset(&decode, 0, sizeof(decode));
    decode.dst_address = dst;

    if(decompress_xz((unsigned char*)src, (cyg_uint32)-1, &decode)) {
        diag_printf("Decompression error: %s\n", decode.err_msg);
    } else {
        load_address     = entry_address = (CYG_ADDRESS)dst;
        load_address_end = load_address + decode.decoded_length;;
    }
}

RedBoot_cmd("xz",
	    "Uncompress XZ compressed data",
	    "-s <location> -d <location>",
	    do_xz
    );
#endif
