//==========================================================================
//
//      vimage.c
//
//      Support code/data for VTSS linux image
//
//==========================================================================
// ####ECOSGPLCOPYRIGHTBEGIN####                                            
// -------------------------------------------                              
// This file is part of eCos, the Embedded Configurable Operating System.   
// Copyright (C) 2006 Free Software Foundation, Inc.                        
//
// eCos is free software; you can redistribute it and/or modify it under    
// the terms of the GNU General Public License as published by the Free     
// Software Foundation; either version 2 or (at your option) any later      
// version.                                                                 
//
// eCos is distributed in the hope that it will be useful, but WITHOUT      
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or    
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License    
// for more details.                                                        
//
// You should have received a copy of the GNU General Public License        
// along with eCos; if not, write to the Free Software Foundation, Inc.,    
// 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.            
//
// As a special exception, if other files instantiate templates or use      
// macros or inline functions from this file, or you compile this file      
// and link it with other works to produce a work based on this file,       
// this file does not by itself cause the resulting work to be covered by   
// the GNU General Public License. However the source code for this file    
// must still be made available in accordance with section (3) of the GNU   
// General Public License v2.                                               
//
// This exception does not invalidate any other reasons why a work based    
// on this file might be covered by the GNU General Public License.         
// -------------------------------------------                              
// ####ECOSGPLCOPYRIGHTEND####                                              
//==========================================================================
//#####DESCRIPTIONBEGIN####
//
// Author(s):    Lars Povlsen
// Contributors: 
// Date:         2015-01-07
// Purpose:      
// Description:  
//              
// This code is part of RedBoot (tm).
//
//####DESCRIPTIONEND####
//
//==========================================================================

#include <redboot.h>

#ifdef CYGPKG_COMPRESS_XZLIB

#include <alloca.h>
#include <vimage.h>
#include <cyg/compress/xz.h>
#include <cyg/crc/crc.h>
#include <cyg/infra/cyg_ass.h>         // assertion macros
#ifdef CYGSEM_REDBOOT_FIRMWARE_AUTHENTICATE
#include <fis.h>
#include <cyg/io/flash.h>
#include <cyg/libtomcrypt/tomcrypt.h>
#endif
#include "mscc_md5_api.h"

#define RL_PUB_SEC_KEY "PUB_SEC_KEY=-----BEGIN PUBLIC KEY-----"
#define RL_KEY_END     "-----END PUBLIC KEY-----"
#define CONF_LENGTH    1024

const cyg_uint8 xz_magic[6] = { 0xFD, '7', 'z', 'X', 'Z', 0x00 };

#if defined(CYGSEM_REDBOOT_FIRMWARE_AUTHENTICATE) && defined(CYG_HAL_STARTUP_ROMRAM)
static bool rot_valid = true;
#endif

int vimage_socfamily_valid(int socfamily)
{
#if defined(CYG_HAL_VCOREIII_CHIPTYPE_LUTON26)
    return socfamily == 2;
#elif defined(CYG_HAL_VCOREIII_CHIPTYPE_JAGUAR)
    return socfamily == 3;
#elif defined(CYG_HAL_VCOREIII_CHIPTYPE_SERVAL)
    return socfamily == 4;
#elif defined(CYG_HAL_VCOREIII_CHIPTYPE_JAGUAR2)
    return socfamily == 5;
#elif defined(CYG_HAL_VCOREIII_CHIPTYPE_SERVAL_T)
    return socfamily == 6;
#elif defined(CYG_HAL_VCOREIII_CHIPTYPE_JAGUAR2_C)
    return socfamily == 7;
#elif defined(CYG_HAL_VCOREIII_CHIPTYPE_OCELOT)
    return socfamily == 8;
#else
 #error Unsupported ecos platform
#endif
}

int vimage_hdrcheck(struct firmware_vimage *fw, const char **errmsg)
{
    if (fw->magic1 != VIMAGE_HDR_MAGIC1 ||
        fw->magic2 != VIMAGE_HDR_MAGIC2) {
        *errmsg = "Image format invalid";
        return -1;
    }
    if (fw->hdrlen < VIMAGE_HDR_MAX) {
        cyg_uint32 calccrc, orgcrc = fw->hdrcrc;
        fw->hdrcrc = 0;
        calccrc = cyg_posix_crc32((unsigned char *)fw, fw->hdrlen);
        fw->hdrcrc = orgcrc;    // Restore CRC state;
        if (calccrc != orgcrc) {
            *errmsg = "Image header CRC invalid";
            return -1;
        }
    } else {
        *errmsg = "Image header length invalid";
        return -1;
    }

    if (fw->kern_len == 0 || fw->kern_off == 0) {
        *errmsg = "No kernel data";
        return -1;
    }

    if (!vimage_socfamily_valid(fw->soc_no)) {
        *errmsg = "Image with wrong chipset family";
        return -1;
    }

    if (fw->kern_cmd_len) {
        if (fw->kern_cmd_len >= VIMAGE_CMD_MAX) {
            *errmsg = "Kernel command line length invalid";
            return -1;
        }
        if (fw->kern_cmd[fw->kern_cmd_len] != '\0') {
            *errmsg = "Kernel command not properly terminated";
            return -1;
        }
    }

    return 0;    // All OK
}

int mscc_vimage_hdrcheck(mscc_firmware_vimage_t *fw, const char **errmsg)
{ 
    if (fw->hdrlen < sizeof(*fw)) {
        *errmsg = "Short header";
        return -1;
    }
    if (fw->magic1 != MSCC_VIMAGE_MAGIC1 ||
        fw->magic2 != MSCC_VIMAGE_MAGIC2 ||
        fw->version < 1) {
        *errmsg = "Image format invalid";
        return -1;
    }
    if (!vimage_socfamily_valid(fw->soc_no)) {
        *errmsg = "Image with wrong chipset family";
        return -1;
    }
    return 0;    // All OK
}

#ifdef CYGSEM_REDBOOT_FIRMWARE_AUTHENTICATE

// RL_MODS - Start
static void load_auth_key(char *fis_key, rsa_key *key)
{
    unsigned char *b64Start = NULL,          // starting point of base64 in "conf" image
                  *b64End = NULL;            // ending point of base64 in "conf" image

    cyg_uint8     conf[CONF_LENGTH],         // Contents of conf partition
                  der[CONF_LENGTH];          // DER from the PEM

    unsigned int  confLen = CONF_LENGTH;     // Bytes to read from "conf" image

    unsigned long derLen = CONF_LENGTH;      // Bytes in the DER

    int           ndx = 0,                   // Index in Base64 section of PEM
                  moveNdx,                   // Index used in moving memory
                  ltcStatus,                 // libtomcrypt return status
                  b64Len;                    // Number of base64 chars in PEM

    struct fis_image_desc *fwkey;
    key->type = -1;

    if ((fwkey = fis_lookup(fis_key, NULL))) {
        if ( fwkey->data_length < confLen )  // Partition is smaller than buffer
        {
            confLen = fwkey->data_length;
        }

        cyg_flash_read(fwkey->flash_base, conf, confLen, NULL);

        conf[confLen - 1] = 0;               // Ensure there is a null terminator

        if ( !(b64Start = (unsigned char *)strstr((char *)conf, RL_PUB_SEC_KEY)) )
        {
            diag_printf("Public Key not found in NOR conf partition\n");
            return;
        }
        else if ( !(b64End = (unsigned char *)strstr((char *)b64Start, RL_KEY_END)) )
        {
            diag_printf("Complete Public Key not found in NOR conf partition\n");
            return;
        }
        else
        {
            b64Start += strlen(RL_PUB_SEC_KEY);
            b64Len = b64End - b64Start;

            while ( ndx < b64Len )           // Clean up non-base64 characters
            {
                if ( isBase64Char(*(b64Start + ndx)) )   // This is a base64 character
                {
                    ndx++;
                }
                else                         // This is not a base64 character
                {
                    for ( moveNdx = ndx;
                          moveNdx < (b64Len - 1);
                          moveNdx++ )
                    {
                        *(b64Start + moveNdx) = *(b64Start + moveNdx + 1);
                    }

                    *(b64Start + b64Len - 1) = '\0';  // Shift in a NULL

                    b64Len--;
                }
            }

            ltcStatus = base64_decode(b64Start,
                                      b64Len,
                                      der,
                                      &derLen);

            if ( ltcStatus != CRYPT_OK )
            {
                diag_printf("Could not decode base64 section of key - %d\n",
                            ltcStatus);

                return;
            }
        }

        ltcStatus = rsa_import(der,         // Import the key
                               derLen,
                               key);

        if ( ltcStatus != CRYPT_OK )
        {
            if (key->type != PK_PUBLIC)
            {
                diag_printf("Invalid firmware authentication key type '%d'.\n",
                            key->type);
            }
            else
            {
                diag_printf("Error reading firmware authentication key - %d\n",
                            ltcStatus);
            }
        }
    }
    else
    {
        diag_printf("No firmware authentication key installed.\n");
    }
}

int isBase64Char(char inChar)
{
    int status = false;                         // Default return value

    if ( ((inChar >= 'A') && (inChar <= 'Z')) ||
         ((inChar >= 'a') && (inChar <= 'z')) ||
         ((inChar >= '0') && (inChar <= '9')) ||
         (inChar == '+') ||
         (inChar == '/') ) status = true;

    return status;
}

// RL_MODS - End

static bool do_sha_validate(mscc_firmware_vimage_t *fw,
                            const char *sha_name, size_t digest_size,
                            const void *signature, cyg_uint32 sig_size)
{
    rsa_key fw_key;
    int hash_idx;
    unsigned long hlen = digest_size;
    cyg_uint8 *digest2;
    load_auth_key("conf", &fw_key);            // RL_MODS - different partition

    int ltcStatus;                             // RL_MODS

    if (fw_key.type != PK_PUBLIC) {
        return false;
    }
    if ((digest2 = alloca(digest_size)) == NULL) {
        diag_printf("Digest allloca fails: %zd words.\n", digest_size);
        return false;
    }
    if ((hash_idx = find_hash(sha_name)) == -1) {
        diag_printf("Hash type %s not available.\n", sha_name);
        return false;
    }
    if (hash_memory(hash_idx,
                    (unsigned char *)fw, fw->imglen,
                    digest2, &hlen) == CRYPT_OK) {
        int stat;

        if ( (ltcStatus = rsa_verify_hash_ex(signature,
                                             sig_size,
                                             digest2,
                                             hlen,
                                             LTC_PKCS_1_V1_5,
                                             hash_idx,
                                             0,
                                             &stat,
                                             &fw_key)) == CRYPT_OK) {
            if (stat == 1) {
                diag_printf("Firmware was rsa-%s authenticated.\n", sha_name);
                return true;
            } else {
                diag_printf("Firmware rsa-%s authentication error.\n", sha_name);
            }
        } else {
            diag_printf("Firmware rsa-%s signature decryption error [%d].\n",
                        sha_name,
                        ltcStatus);
        }
    } else {
        diag_printf("Hash function %s error.\n", sha_name);
    }
    return false;
}

#endif /* CYGSEM_REDBOOT_FIRMWARE_AUTHENTICATE */

bool mscc_vimage_validate(mscc_firmware_vimage_t *fw, mscc_firmware_vimage_tlv_t *tlv, u8 *signature)
{
    u8 *sigcopy;

    // check signature sanity
    switch (fw->img_sign_type) {
        case MSCC_FIRMWARE_IMAGE_SIGNATURE_MD5:
            if (tlv->data_len != MD5_MAC_LEN)
                goto invalid_signature;
            break;
        case MSCC_FIRMWARE_IMAGE_SIGNATURE_SHA256:
            if (tlv->data_len < 256 || tlv->data_len > 1024)
                goto invalid_signature;
            break;
        case MSCC_FIRMWARE_IMAGE_SIGNATURE_SHA512:
            if (tlv->data_len < 256 || tlv->data_len > 1024)
                goto invalid_signature;
            break;
        case MSCC_FIRMWARE_IMAGE_SIGNATURE_NULL:
        default:
        invalid_signature:
            diag_printf("Image has valid signature");
            return false;
    }

    // Allocate signature copy while re-calculating
    sigcopy = alloca(tlv->data_len);
    if (!sigcopy) {
        return false;    // Failure
    }
    // Copy orig signature
    memcpy(sigcopy, signature, tlv->data_len);
    // Reset to zero
    memset(signature, 0, tlv->data_len);

    switch (fw->img_sign_type) {
//#if !defined(CYGSEM_REDBOOT_FIRMWARE_AUTHENTICATE)
//        case MSCC_FIRMWARE_IMAGE_SIGNATURE_MD5:
//            {
//                struct mscc_MD5Context ctx;
//                mscc_MD5Init(&ctx);
//                mscc_MD5Update(&ctx, (void*)fw, fw->imglen);
//                mscc_MD5Final(signature, &ctx);
//                if (memcmp(sigcopy, signature, tlv->data_len) == 0) {
//                    diag_printf("MD5 signature validated\n");
//                    return true;
//                }
//            }
//            break;
//#else /* defined(CYGSEM_REDBOOT_FIRMWARE_AUTHENTICATE */
        case MSCC_FIRMWARE_IMAGE_SIGNATURE_SHA256:
            {
                if (do_sha_validate(fw, "sha256", 32, sigcopy, tlv->data_len)) {
                    memcpy(signature, sigcopy, tlv->data_len);    // Restore signature
                    return true;
                }
            }
            break;
        case MSCC_FIRMWARE_IMAGE_SIGNATURE_SHA512:
            {
                if (do_sha_validate(fw, "sha512", 64, sigcopy, tlv->data_len)) {
                    memcpy(signature, sigcopy, tlv->data_len);    // Restore signature
                    return true;
                }
            }
            break;
//#endif
        default:
#if defined(CYGSEM_REDBOOT_FIRMWARE_AUTHENTICATE)
            diag_printf("Type %d signature is unsupported\n", fw->img_sign_type);
#else
            diag_printf("Type %d signature skipped\n", fw->img_sign_type);
            return true;    // Skip auth if we shouldn't enforce
#endif
    }
    diag_printf("*** Failed to validate type %d signature ***\n", fw->img_sign_type);
    return false;    // Failed auth
}

u8 *mscc_vimage_find_tlv(const mscc_firmware_vimage_t *fw,
                         mscc_firmware_vimage_tlv_t *tlv,
                         mscc_firmware_image_stage1_tlv_t type)
{
    size_t off = sizeof(*fw);
    mscc_firmware_vimage_tlv_t *wtlv;

    while (off < fw->imglen) {
        wtlv = (mscc_firmware_vimage_tlv_t*) (((u8*)fw) + off);
        if (type == wtlv->type) {
            *tlv = *wtlv;
            return &wtlv->value[0];
        }
        off += wtlv->tlv_len;
    }

    return NULL;
}

int xz_decompress_simple(const cyg_uint8 *src_address, cyg_uint32 src_len, 
                         cyg_uint8 *dst_address, cyg_uint32 dst_len,
                         cyg_uint32 *decoded_length)
{
    struct xz_buf b;
    struct xz_dec *s;
    enum xz_ret ret;

#if 0
    diag_printf("XZ: src %p len %zd dst %p len %zd\n", src_address, src_len, dst_address, dst_len);
#endif

    b.out      = dst_address;
    b.out_size = (size_t) dst_len;
    b.out_pos  = 0;
    b.in       = src_address;
    b.in_size  = (size_t) src_len;
    b.in_pos   = 0;

    s = xz_dec_init(XZ_SINGLE, 0);
    ret = xz_dec_run(s, &b);
    xz_dec_end(s);

    if (ret == XZ_STREAM_END) {
        *decoded_length = b.out_pos;
    }
    
    return ret == XZ_STREAM_END ? 0 : ret;
}

const char *xz_errmsg(int ret)
{
    switch(ret) {
        case XZ_OK:
            return "OK";
        case XZ_STREAM_END:
            return "Stream end (OK)";
        case XZ_UNSUPPORTED_CHECK:
            return "Integrity check type is not supported";
        case XZ_MEM_ERROR:
            return "Allocating memory failed";
        case XZ_MEMLIMIT_ERROR:
            return "Too big LZMA2 dictionary";
        case XZ_FORMAT_ERROR:
            return "File format was not recognized";
        case XZ_OPTIONS_ERROR:
            return "Unsupported options requested";
        case XZ_DATA_ERROR:
            return "Compressed data is corrupt";
        case XZ_BUF_ERROR:
            return "Cannot make any progress";
        default:
            return "Unknown error";
    }
}

void mscc_vimage_rot_set_invalid(void)
{
#if defined(CYGSEM_REDBOOT_FIRMWARE_AUTHENTICATE) && defined(CYG_HAL_STARTUP_ROMRAM)
    rot_valid = false;
#endif
}

#endif    // CYGPKG_COMPRESS_XZLIB
